@@ -95,6 +95,18 @@ pts= np.random.random((3,3))
 
 from math import tanh
 
+def get_radius(curve):
+    num = curve.sample_size
+    rad = []
+    for i in np.linspace(0,1,num=num):
+        _,d,dd = curve.derivatives(i,2)
+        d  = np.array(d)
+        dd = np.array(dd)
+        value = np.linalg.norm(d)**3/(np.linalg.norm(np.cross(d,dd))+np.finfo(float).eps)
+        rad.append(value)
+    return rad
+
+
 @nb.jit(nopython=True)
 def get_collisions(collision_vessels,R,sample_pts):
     collisions = 0
@@ -105,12 +117,22 @@ def get_collisions(collision_vessels,R,sample_pts):
     return collisions
 
 
+@nb.jit(nopython=True)
+def are_collisions(collision_vessels,R,sample_pts):
+    collisions = False
+    for i in range(sample_pts.shape[0]):
+        dist = close_exact(collision_vessels,sample_pts[i,:])
+        if np.any(dist<R):
+            collisions = True
+            break
+    return collisions
+
+
 @nb.jit(nopython=True)
 def close_exact(data,point):
     line_direction = np.zeros((data.shape[0],3))
     ss = np.zeros(data.shape[0])
     tt = np.zeros(data.shape[0])
-    #decision = np.zeros((data.shape[0]))
     hh = np.zeros(data.shape[0])
     cc = np.zeros((data.shape[0],3))
     cd = np.zeros(data.shape[0])
@@ -127,6 +149,153 @@ def close_exact(data,point):
         line_distances[i] = np.sqrt(hh[i]**2+cd[i]**2) - data[i,6]
     return line_distances
 
+def connect_deterministic(P1,P2,P3,P4,R,collision_vessels):
+    MID = (P2 + P4)/2
+    P2_P4_to_MID_dist = np.linalg.norm(P2 - MID)
+    V1 = P2 - P1
+    V1 = V1/np.linalg.norm(V1)
+    V2 = P4 - P3
+    V2 = V2/np.linalg.norm(V2)
+    success = False
+    points = [P2]
+    vectors = [V1]
+    while not success:
+        # Check direct connection
+        check = P4 - points[-1]
+        check = check/np.linalg.norm(check)
+        vectors_check = vectors
+        vectors_check.append(check)
+        vectors_check.append(V2)
+        vectors_check = np.array(vectors_check)
+
+
+from geomdl import BSpline, utilities
+from geomdl.visualization import VisMPL
+
+
+def connect_bezier(P1,P2,P3,P4,R,collision_vessels,degree=2):
+    V1   = (P2 - P1)
+    V1   = V1/np.linalg.norm(V1)
+    V2   = (P4 - P3)
+    V2   = V2/np.linalg.norm(V2)
+    def create_bezier(data,V1=V1,V2=V2,R=R,P2=P2,P4=P4,degree=degree):
+        CTR0 = P2 + (R+data[0])*V1
+        CTR1 = P4 + (R+data[1])*V2
+        CTR  = np.zeros(data.shape[0]-2+12)
+        fill_CTR = np.zeros(data.shape[0]-2).reshape(-1,3)
+        seg = np.linspace(0.1,0.9,num=fill_CTR.shape[0])
+        for i in range(fill_CTR.shape[0]):
+            fill_CTR[i,:] = CTR0*(1-seg[i]) + CTR1*(seg[i]) #V1*(1-seg[i])*R + CTR1*(seg[i]) + V2*(seg[i])*R
+        fill_CTR = fill_CTR.flatten()
+        CTR[0:3]     = P2
+        CTR[3:6]     = CTR0
+        CTR[-6:-3]   = CTR1
+        CTR[-3:]     = P4
+        CTR[6:-6]    += fill_CTR + data[2:]
+        CTR  = CTR.reshape(-1,3).tolist()
+        curve = BSpline.Curve()
+        curve.degree = degree
+        curve.ctrlpts = CTR
+        curve.knotvector = utilities.generate_knot_vector(curve.degree,len(curve.ctrlpts))
+        curve.sample_size = 40
+        curve.evaluate()
+        return curve
+    return create_bezier
+
+
+def bezier_cost(data,grad,create_curve=None,R=None,P1=None,P3=None,collision_vessels=None):
+    curve = create_curve(data)
+    pts   = np.array(curve.evalpts)
+    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
+    #num = int(spline_length // R)
+    #curve.sample_size = num
+    #curve.evaluate()
+    sample_pts = np.array(curve.evalpts)
+    sample_pts_for_vec = np.vstack((P1,sample_pts,P3))
+    vec = np.diff(sample_pts_for_vec,axis=0)
+    collisions = get_collisions(collision_vessels,R,sample_pts)
+    #vec = vec/np.linalg.norm(vec,axis=0)
+    #angles = get_vecs(vec)
+    curve_rads = np.array(get_radius(curve))
+    curve_rads = 1/(curve_rads[curve_rads<R]+np.finfo(float).eps)
+    a_sum = np.sum(curve_rads)
+    if np.isclose(a_sum,0):
+        vec = vec/np.linalg.norm(vec,axis=0)
+        angles = 1/get_vecs(vec)
+        if np.any(angles < 120):
+            a_sum = np.sum(1/angles)
+    return collisions+a_sum*spline_length
+
+
+def test_b(P1,P2,P3,P4,R,collision_vessels):
+    create_curve = connect_bezier(P1,P2,P3,P4,R,collision_vessels)
+    cost = lambda d,g: bezier_cost(d,g,create_curve=create_curve,R=R,P1=P1,P3=P3,collision_vessels=collision_vessels)
+    vis_comp = VisMPL.VisCurve3D()
+    success = False
+    x0 = np.zeros(11)
+    max_time = 25
+    count = 1
+    L = 4*R
+    level_attempts = 4
+    attempt = 0
+    best = np.inf
+    while not success:
+        print('Linking Optimization Path Search {} Level {}'.format(count,attempt))
+        lb = np.ones((count+2)*3+2)*(-L) + x0
+        lb[0] = 0
+        lb[1] = 0
+        ub = np.ones((count+2)*3+2)*(L) + x0
+        o = nlopt.opt(2,(count+2)*3+2)
+        o.set_min_objective(cost)
+        o.set_lower_bounds(lb)
+        o.set_upper_bounds(ub)
+        o.set_maxtime(max_time)
+        xopt = o.optimize(x0)
+        curve = create_curve(xopt)
+        curve.vis = vis_comp
+        if o.last_optimum_value() < 100 and attempt == level_attempts:
+            L += L*0.25
+            success = True
+            print(o.last_optimum_value())
+        elif attempt < level_attempts:
+            attempt += 1
+            max_time += 5
+            if o.last_optimum_value() < best:
+                x0 = xopt
+                best = o.last_optimum_value()
+                curve.render()
+            if np.isclose(o.last_optimum_value(),0):
+                success = True
+                print('Constraints Satisfied: Exiting...')
+            continue
+        else:
+            print("Above link threshold {}".format(o.last_optimum_value()))
+            L += L*0.25
+            count += 1
+            attempt = 0
+            max_time = 25
+            x0 = np.zeros((count+2)*3+2)
+    curve = create_curve(xopt)
+    return curve
+
+
+def get_connect_points(P1,P2,P3,P4,R,collision_vessels):
+    curve = test_b(P1,P2,P3,P4,R,collision_vessels)
+    pts   = np.array(curve.evalpts)
+    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
+    num = int(spline_length // R)+2
+    curve.sample_size = num
+    curve.evaluate()
+    opt_pts = np.array(curve.evalpts)
+    fig = plt.figure()
+    ax  = fig.add_subplot(111,projection='3d')
+    ax.plot(opt_pts[:,0],opt_pts[:,1],opt_pts[:,2],label='final')
+    plt.show()
+    return opt_pts
+
+def connect_linear_search(P1,P2,P3,P4,R,collsiion_vessels):
+    pass
+
 def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
     #P1 = np.random.random(3)
     #P2 = np.random.random(3)
@@ -184,12 +353,12 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
         CTR = CTR + data.reshape(-1,3)
         SPT1 = np.vstack((P2,CTR0,CTR,CTR4,P4))
         try:
-            tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=2)
+            tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=3)
         except:
             CTR1 = CTR1 - data[0:3]
             CTR2 = CTR2 - data[3:]
             SPT1 = np.vstack((P2,CTR0,CTR1,CTR2,CTR4,P4))
-            tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=2)
+            tck1, u = interpolate.splprep([SPT1[:,0],SPT1[:,1],SPT1[:,2]],s=0,k=3)
         spline_f1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=0))
         spline_df1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=1))
         spline_ddf1 = lambda t: np.array(interpolate.splev(t,tck=tck1,der=2))
@@ -214,7 +383,7 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
     init = np.zeros(9)
     #init[0:3] = CTR1
     #init[3:]  = CTR2
-    def cost(data,grad,spline_gen=get_spline,angle_threshold=angle_threshold,R=R,P1=P1,CTR4=CTR4,collision_vessels=collision_vessels):
+    def cost(data,grad,spline_gen=get_spline,angle_threshold=angle_threshold,R=R,P1=P1,P2=P2,P4=P4,CTR4=CTR4,collision_vessels=collision_vessels):
         spf,spdf,spddf = get_spline(data)
         tt = np.linspace(0,1,num=500)
         pts = spf(tt).T
@@ -222,8 +391,8 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
         num = int(spline_length // R) #(R*2*np.pi))
         sample_tt = np.linspace(0,1,num=4*num+2)
         sample_pts = spf(sample_tt).T
-        sample_pts = np.vstack((P1,sample_pts,CTR4))
-        vec = np.diff(sample_pts,axis=0)
+        sample_pts_for_vec = np.vstack((P1,P2,sample_pts,P4,P3))
+        vec = np.diff(sample_pts_for_vec,axis=0)
         collisions = get_collisions(collision_vessels,R,sample_pts)
         vec = vec/np.linalg.norm(vec,axis=0)
         angles = []
@@ -231,7 +400,7 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
         angles = get_vecs(vec)
         a_sum = np.exp(np.sum(angles))
         a_try = np.sum(angles)
-        return collisions+a_sum*spline_length #a_sum+spline_length + collisions#+np.tanh(num)
+        return collisions+a_sum+spline_length
     def cost_polish(data,grad,spline_gen=get_spline,angle_threshold=angle_threshold,R=R,P1=P1,CTR4=CTR4,collision_vessels=collision_vessels):
         spf,spdf,spddf = get_spline(data)
         tt = np.linspace(0,1,num=500)
@@ -283,12 +452,13 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
         spf,spdf,spddf = get_spline(data)
         curve = lambda t: curvature(t,df=spdf,ddf=spddf)
         return curve
-    def cost2(data,R=R,get_spline=get_spline):
+    def cost2(data,grad,R=R,get_spline=get_spline,collision_vessels=collision_vessels):
         spf,spdf,spddf = get_spline(data)
         tt = np.linspace(0,1,num=500)
         pts = spf(tt).T
         spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
         kappa = get_curvature(data)
+        collisions = get_collisions(collision_vessels,R,pts)
         tt = np.linspace(0,1,num=100)
         k = []
         for t in tt:
@@ -297,12 +467,12 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
                 k.append(1)
             else:
                 k.append(0)
-        return np.sum(k)*spline_length
+        return collisions + np.sum(k)*spline_length
     def cost3(t_vector,grad,spline=None,angle_threshold=angle_threshold,P1=P1,P2=P2,P3=P3,MID=MID,CTR4=CTR4):
         tt = np.linspace(0,1,num=len(t_vector)+2)
         tt[1:-1] = t_vector
         sample_pts = spline(t_vector).T
-        sample_pts = np.vstack((P1,sample_pts,P3))
+        sample_pts = np.vstack((P1,P2,sample_pts,P4,P3))
         vec = np.diff(sample_pts,axis=0)
         vec = vec/np.linalg.norm(vec,axis=0)
         angles = []
@@ -314,7 +484,7 @@ def opt(P1,P2,P3,P4,R,collision_vessels): #P1,P2,P3,P4,R):
                 angles.append(0)
         a_sum = np.exp(100*np.sum(angles))
         return a_sum
-    return get_spline,show,cost,init,LENGTH,check,get_points,get_curvature,cost3,collision_vessels,cost_polish
+    return get_spline,show,cost2,init,LENGTH,check,get_points,get_curvature,cost3,collision_vessels,cost_polish
 
 
 def form_connection(R):
@@ -325,8 +495,8 @@ def form_connection(R):
     pts = gp(res1.x)
     return pts
 
-def test_nlopt(R):
-    f,s,c,i,L,ch,gp,gc,c3,cv,cp = opt(R)
+def test_nlopt(P1,P2,P3,P4,R,collision_vessels):
+    f,s,c,i,L,ch,gp,gc,c3,cv,cp = opt(P1,P2,P3,P4,R,collision_vessels)
     fig = plt.figure()
     ax1 = fig.add_subplot(111,projection='3d')
     #ax2 = fig.add_subplot(122,projection='3d')
@@ -404,18 +574,18 @@ def test_nlopt(R):
     lb = np.array(lb)
     ub = np.array(ub)
     t0 = (lb + ub)/2
-    print("Performing Vessel Annealing Optimization")
+    #print("Performing Vessel Annealing Optimization")
     seg_opt = nlopt.opt(2,len(t0))
     seg_opt.set_min_objective(cost3)
     seg_opt.set_lower_bounds(lb)
     seg_opt.set_upper_bounds(ub)
     seg_opt.set_stopval(100)
     seg_opt.set_maxtime(100)
-    t_vector = seg_opt.optimize(t0)
-    tt = np.linspace(0,1,num=len(t_vector)+2)
-    tt[1:-1] = t_vector
-    spline1 = spf(tt).T
-    ax1.plot(spline1[:,0],spline1[:,1],spline1[:,2],color='blue',label='final')
+    #t_vector = seg_opt.optimize(t0)
+    #tt = np.linspace(0,1,num=len(t_vector)+2)
+    #tt[1:-1] = t_vector
+    #spline1 = spf(tt).T
+    #ax1.plot(spline1[:,0],spline1[:,1],spline1[:,2],color='blue',label='final')
     plt.show()
     return xopt,spf
 
@@ -524,8 +694,7 @@ def retest(data,f,s,c,i,L,ch):
     print('Min angle:\n{}'.format(ANGLES[1]))
     print('Max angle:\n{}'.format(ANGLES[0]))
 
-def find(spf):
-    inital_length
+
 
 def show(spl):
     fig = plt.figure()
@@ -538,3 +707,63 @@ def show(spl):
     #ax.scatter(SPT1[:,0],SPT1[:,1],SPT1[:,2],color='black',label='data')
     #ax.legend()
     plt.show()
+
+
+##
+# Test 1
+##
+
+P1 = np.array([0.5,0,0.5])
+P2 = np.array([0.5,0.5,0.5])
+P3 = np.array([0.5,1,0])
+P4 = np.array([0.5,0.5,0])
+R  = 0.1
+collision_vessels = np.array([[1,0,0,1,0,1,0.1]])
+c = test_b(P1,P2,P3,P4,R,collision_vessels)
+
+
+P1 = np.array([0.5,0,0.5])
+P2 = np.array([0.5,0.5,0.5])
+P3 = np.array([0.5,0,0])
+P4 = np.array([0.5,0.5,0])
+R  = 0.1
+collision_vessels = np.array([[1,0,0,1,0,1,0.1]])
+c = test_b(P1,P2,P3,P4,R,collision_vessels)
+
+P1 = np.array([0.5,0,0.5])
+P2 = np.array([0,0.5,0.3])
+P3 = np.array([0.5,1,0])
+P4 = np.array([0.5,0.5,0])
+R  = 0.1
+collision_vessels = np.array([[1,0,0,1,0,1,0.1]])
+c = test_b(P1,P2,P3,P4,R,collision_vessels)
+
+P1 = np.array([0.5,0,0.5])
+P2 = np.array([0,0.5,0.3])
+P3 = np.array([1,0,0])
+P4 = np.array([1,0.5,0])
+R  = 0.1
+collision_vessels = np.array([[1,0,0,1,0,1,0.1]])
+c = test_b(P1,P2,P3,P4,R,collision_vessels)
+
+P1 = np.array([0.5,0,0.5])
+P2 = np.array([0.5,0.5,0.5])
+P3 = np.array([0.5,1,0])
+P4 = np.array([0.5,0.5,0])
+R  = 0.1
+collision_vessels = np.array([[0,0.5,0.25,1,0.5,0.25,0.01],
+                              [0,0.7,0.25,1,0.7,0.25,0.01],
+                              [0,0.3,0.25,1,0.3,0.25,0.01],
+                              [0,0.9,0.25,1,0.9,0.25,0.01],
+                              [0,0.1,0.25,1,0.1,0.25,0.01],
+                              [0.25,0,0.25,0.25,1,0.25,0.01],
+                              [0.75,0,0.25,0.75,1,0.25,0.01],
+                              [0,0.5,0.75,1,0.5,0.75,0.01],
+                              [0,0.7,0.75,1,0.7,0.75,0.01],
+                              [0,0.3,0.75,1,0.3,0.75,0.01],
+                              [0,0.9,0.75,1,0.9,0.75,0.01],
+                              [0,0.1,0.75,1,0.1,0.75,0.01],
+                              [0.25,0,0.75,0.25,1,0.75,0.01],
+                              [0.75,0,0.75,0.75,1,0.75,0.01],
+                              [1,0,0,1,0,1,0.1]])
+c = test_b(P1,P2,P3,P4,R,collision_vessels)
