@@ -4,7 +4,8 @@ from geomdl import utilities
 from geomdl.visualization import VisMPL
 import numpy as np
 from copy import deepcopy
-
+from .optimize_connection_v2 import *
+from tqdm import trange
 
 def get_initial_vectors(forest,network,assignment,tree,edge):
     # Terminal Vessel Vector
@@ -20,33 +21,46 @@ def get_initial_vectors(forest,network,assignment,tree,edge):
     V2 = V2/np.linalg.norm(V2)
     return V1,V2,P4
 
-def get_angle(V1,V2):
-    return np.arccos(np.dot(-V1,V2))*(180/np.pi)
 
-
-def link(forest,angle_threshold=120):
+def link(forest):
     all_connections = []
     for network in range(forest.number_of_networks):
         network_connections = []
+        all_network_vessels = np.zeros((1,7))
+        for tree in range(forest.trees_per_network[network]):
+            tree_tmp_vessels = np.zeros((forest.networks[network][tree].data.shape[0],7))
+            tree_tmp_vessels[:,0:6] = forest.networks[network][tree].data[:,0:6]
+            tree_tmp_vessels[:,6]   = forest.networks[network][tree].data[:,21]
+            all_network_vessels = np.vstack((all_network_vessels,tree_tmp_vessels))
+        all_network_vessels = all_network_vessels[1:,:]
+        all_network_mid_points = (all_network_vessels[0:3]+all_network_vessels[3:6])/2
         for tree in range(forest.trees_per_network[network]):
             assignment = forest.assignments[network][tree]
             tree_connections = []
-            for edge in range(len(assignment)):
+            for edge in trange(len(assignment)):
                 R        = forest.networks[network][tree].data[assignment[edge],21]
-                V1,V2,P4 = get_initial_vectors(forest,network,assignment,tree,edge)
-                angle    = get_angle(V1,V2)
-                success  = False
-                temp_end = P4
-                while not success:
-                    if angle > angle_threshold:
-                        # check new vessel for collision
-                        if not collision:
-                            # append vessel to link list
-                            if np.all(temp_end==P4):
-                                # if the new vessel reaches the connection
-                                # endpoint then stop iterating on vessel
-                                # creation
-                                success = True
+                P1       = forest.networks[network][tree].data[assignment[edge],0:3]
+                P2       = forest.networks[network][tree].data[assignment[edge],3:6]
+                ignore   = set(np.argwhere(np.all(all_network_mid_points[3:6]==P2)).flatten().tolist())
+
+                P4       = forest.connections[network][0][edge]
+                V        = P4 - P2/np.linalg.norm(P4-P2)
+                P3       = P4 + V*0.25
+                dist_to_mid = np.linalg.norm(P4-P2)
+                dists       = np.linalg.norm(all_network)
+                collision_vessels_idx = set(np.argwhere(dists<dist_to_mid).flatten().tolist())
+                if len(collision_vessels_idx) == 0:
+                    collision_vessels = None
+                else:
+                    collision_vessels_idx = np.array(list(collision_vessels_idx - ignore))
+                    collision_vessels = all_network_vessels[collision_vessels_idx,:]
+                link_pts = get_optimum_link_points(P1,P2,P3,P4,R,collision_vessels)
+                link_pts = link_pts[1:-1,:]
+                tree_connections.append(link_pts)
+            network_connections.append(tree_connections)
+        all_connections.append(network_connections)
+    return all_connections
+
 def smooth_original(forest,curve_sample_size_min=5,curve_sample_size_max=21,curve_degree=3):
     #FOREST_CONNECT_COPY = deepcopy(forest)
     connections = []
